#include <stdio.h>
#include <stdlib.h>
#include <usb.h>

static struct usb_device *probe_for_device(u_int16_t vendor, u_int16_t product)
{
  struct usb_bus *busses;
  struct usb_bus *bus;
  struct usb_device *dev;
  
  usb_find_busses();
  usb_find_devices();
  busses = usb_get_busses();
  for (bus = busses; bus; bus = bus->next) {
    for (dev = bus->devices; dev; dev = dev->next) {
      /* Check if this device is the one we're looking for */
      if((dev->descriptor.idVendor == vendor) && (dev->descriptor.idProduct == product)) return dev;
    }
  }
  return NULL;
}

#define BULK_TIMEOUT 1000

#define CFGIO 0x0d /* See http://jbnote.free.fr/prism54usb/usb_protocol_init.html */

#define EECTL 0x04

static int write_eectl_register(usb_dev_handle *device, u_int32_t data)
{
  char buffer[10];
  
  buffer[0] =  0x1F; /* 00 01 1111 */
  buffer[1] = 0;
  buffer[2] = EECTL;
  buffer[3] = 0;
  buffer[4] = 0;
  buffer[5] = 0;
  buffer[6] = data & 0x000000ff;
  buffer[7] = (data & 0x0000ff00) >> 8;
  buffer[8] = (data & 0x00ff0000) >> 16;
  buffer[9] = (data & 0xff000000) >> 24;
  return(usb_bulk_write(device, CFGIO, &buffer[0], 10, BULK_TIMEOUT) >= 0);
}

static int read_eectl_register(usb_dev_handle *device, u_int32_t *data)
{
  char buffer[6];
  int err;
  
  buffer[0] =  0x1F; /* 00 01 1111 */
  buffer[1] = 0;
  buffer[2] = EECTL;
  buffer[3] = 0;
  buffer[4] = 0;
  buffer[5] = 0;
  err = usb_bulk_write(device, CFGIO, &buffer[0], 6, BULK_TIMEOUT);
  if(err < 0) return err;
  return(usb_bulk_read(device, CFGIO, (char *)data, 4, BULK_TIMEOUT) >= 0);
}

#define EEPROM_BUSY 19
#define EEPROM_CS_ENABLE 18
#define EEPROM_BYTE_READ_START 17
#define EEPROM_BYTE_WRITE_START 16
#define EEPROM_READ_DATA 8
#define EEPROM_WRITE_DATA 0
#define EEPROM_ADDRESS_WIDTH 23

#define READ_EE_OPCODE 3

static int eeprom_wait(usb_dev_handle *device, u_int32_t *regval)
{
  u_int32_t data;
  
  do {
    if(!read_eectl_register(device, &data)) return 0;
  } while((data & (1 << EEPROM_BUSY)) != 0);
  if(regval != NULL) *regval = data;
  return 1;
}

static int eeprom_off(usb_dev_handle *device)
{
  if(!eeprom_wait(device, NULL)) return 0;
  return(write_eectl_register(device, 0));
}

static int eeprom_readbyte(usb_dev_handle *device, char *data)
{
  u_int32_t eectl;
  
  if(!eeprom_wait(device, &eectl)) return 0;
  eectl |= ((1 << EEPROM_CS_ENABLE) | (1 << EEPROM_BYTE_READ_START));
  if(!write_eectl_register(device, eectl)) return 0;
  if(!eeprom_wait(device, &eectl)) return 0;
  *data = (eectl >> EEPROM_READ_DATA) & 0xff;
  return 1;
}

static int eeprom_writebyte(usb_dev_handle *device, char data)
{
  u_int32_t eectl;
  
  if(!eeprom_wait(device, &eectl)) return 0;
  eectl &= ~(0xff << EEPROM_WRITE_DATA);
  eectl |= (1 << EEPROM_CS_ENABLE) | (1 << EEPROM_BYTE_WRITE_START) | (data << EEPROM_WRITE_DATA);
  return(write_eectl_register(device, eectl));
}

static int eeprom_read(usb_dev_handle *device, int startaddr, int threebyte, char *buffer, int size)
{
  int i;

  eeprom_writebyte(device, READ_EE_OPCODE);
  if(threebyte) eeprom_writebyte(device, (startaddr >> 16) & 0xff);
  eeprom_writebyte(device, (startaddr >> 8) & 0xff);
  eeprom_writebyte(device, startaddr & 0xff);
  for(i=0;i<size;i++) {
    if(!eeprom_readbyte(device, buffer+i)) return 0;
    printf("read %d bytes\r", i+1);
  }
  puts("");
  eeprom_off(device);
  return 1;
}

static int eeprom_get_address_width(usb_dev_handle *device)
{
  u_int32_t eectl;
  
  if(!read_eectl_register(device, &eectl)) return 0;
  return((eectl >> EEPROM_ADDRESS_WIDTH) & 0xff);
}

int main(int argc, char *argv[])
{
  struct usb_device *dev;
  usb_dev_handle *devhandle;
  u_int16_t vendor, product;
  char *c;
  int i;
  int size, threebyte;
  char buffer[8192];
  FILE *fd;
  
  if(argc != 4) {
    puts("net2280dump - dumps net2280-based USB adapters EEPROMs");
    puts("by Sebastien BOURDEAUDUCQ - http://yoshiyo.ath.cx/seb - sebastien.b@swissinfo.org");
    puts("Usage: net2280dump <vendorid> <productid> <destfile>");
    return 1;
  }
  vendor = strtoul(argv[1], &c, 0);
  if(*c != 0) {
    puts("Invalid specified vendor ID");
    return 1;
  }
  product = strtoul(argv[2], &c, 0);
  if(*c != 0) {
    puts("Invalid specified product ID");
    return 1;
  }
  usb_init();
  dev = probe_for_device(vendor, product);
  if(dev == NULL) {
    puts("Error - device not found");
    return 1;
  }
  puts("Device found");
  devhandle = usb_open(dev);
  if(usb_claim_interface(devhandle, 0) < 0) {
    puts("Couldn't claim interface");
    usb_close(devhandle);
    return 1;
  }
  
  size = 255; /* To avoid compiler warnings */
  threebyte = 0;
  
  i = eeprom_get_address_width(devhandle);
  switch(i) {
    case 0:
      puts("Failed to get EEPROM address width");
      usb_release_interface(devhandle, 0);
      usb_close(devhandle);
      return 1;
    case 1:
      puts("EEPROM found (using 1-byte address width)");
      size = 255;
      threebyte = 0;
      break;
    case 2:
      puts("EEPROM found (using 2-bytes address width)");
      size = 8192;
      threebyte = 0;
      break;
    case 3:
      puts("EEPROM found (using 3-bytes address width)");
      size = 8192;
      threebyte = 1;
      break;
    default:
      printf("eeprom_get_address_width() returned %d\n", i);
      usb_release_interface(devhandle, 0);
      usb_close(devhandle);
      return 1;
  }

  if(!eeprom_read(devhandle, 0, threebyte, &buffer[0], size)) {
    puts("Failed to read EEPROM");
    usb_release_interface(devhandle, 0);
    usb_close(devhandle);
    return 1;
  }
  
  usb_release_interface(devhandle, 0);
  usb_close(devhandle);

  /* Save the contents of the EEPROM to file */
  fd = fopen(argv[3], "wb");
  if(fd == NULL) {
    puts("Failed to open file for writing");
    return 1;
  }
  if(fwrite(&buffer[0], size, 1, fd) != 1) {
    puts("Failed to save data to file");
    fclose(fd);
    return 1;
  }
  fclose(fd);
  printf("%d bytes from the EEPROM successfully dumped to %s\n", size, argv[3]);

  return 0;
}
